from ruleau import rule


@rule(rule_id="rul_01", name="is_registered")
def is_registered(context, payload):
    """
    Check person is registered
    >>> is_registered(None, {"registered": True})
    True
    >>> is_registered(None, {"registered": False})
    False
    """
    return "registered" in payload and payload["registered"]



import collections
from typing import Dict, List


# There is an issue where test coverage can't unpack next iteration
def calculate_rules_depends_on(rules: List) -> List:  # pragma: no cover
    """
    Calculate rules depends_on using dependencies of parent rules
    """
    for rule in rules:
        for dependency in rule["dependencies"]:
            dependency_list_position = next(
                (
                    i
                    for i, rule_dict in enumerate(rules)
                    if rule_dict["id"] == dependency["id"]
                )
            )
            rules[dependency_list_position]["depends_on"].append(rule["id"])
    return rules


def calculate_rules_breadcrumbs(rules_dict: Dict[str, Dict]) -> Dict[str, Dict]:
    """
    Calculate rule breadcrumb trail to top level rule
    """
    for rule_id in rules_dict:
        depends_on_dict = calculate_rule_depends_on_hierarchies(rule_id, rules_dict)
        depends_on_breadcrumbs = flatten_rule_depends_on_hierarchies(depends_on_dict)
        rules_dict[rule_id]["depends_on_breadcrumbs"] = depends_on_breadcrumbs

    return rules_dict


def calculate_rule_depends_on_hierarchies(
    rule_id: str, rules_dict: Dict[str, Dict]
) -> Dict[str, Dict]:
    """
    Extract hierarchy of depends_on for a given rule
    """
    rule = rules_dict[rule_id]
    depends_on_dict = {}
    for depends_on_id in rule["depends_on"]:
        depends_on_dict[depends_on_id] = calculate_rule_depends_on_hierarchies(
            depends_on_id, rules_dict
        )

    return depends_on_dict


def flatten_rule_depends_on_hierarchies(depends_on_dict: Dict) -> List:
    """
    Flatten hierarchy of depends_on for a given rule
    Creates multiple hierarchies if one-to-many/many-to-one relationships between rules
    """
    depends_on_lists = convert_rule_depends_on_dict_to_lists(depends_on_dict)
    flattened_lists = []

    if depends_on_lists:
        for depends_on_list in depends_on_lists:
            flattened_list = list(flatten_rule_hierarchy(depends_on_list))
            reversed_list = flattened_list[::-1]
            flattened_lists.append(reversed_list)

    return flattened_lists


def convert_rule_depends_on_dict_to_lists(depends_on_dict: Dict) -> List:
    """
    Convert depends_on dictionaries to nested lists structure
    """
    list_collector = []

    for key in depends_on_dict.keys():
        nested_items = convert_rule_depends_on_dict_to_lists(depends_on_dict[key])
        if nested_items:
            for nested_item in nested_items:
                unpacking_list = [key, nested_item]
                list_collector.append(unpacking_list)
        else:
            unpacking_list = [key]
            list_collector.append(unpacking_list)

    if list_collector:
        return list_collector


def flatten_rule_hierarchy(hierarchy: List[List]) -> List:
    """
    Flattens nested lists
    """
    for nested_rule in hierarchy:
        if isinstance(nested_rule, collections.abc.Iterable) and not isinstance(
            nested_rule, (str, bytes)
        ):
            yield from flatten_rule_hierarchy(nested_rule)
        else:
            yield nested_rule


def enrich_rules_doctests_metadata(rules_dict: Dict[str, Dict]) -> Dict[str, Dict]:
    """
    Order and index rules doctests (order: True rules -> False rules)
    """
    for rule_id in rules_dict:
        rules_dict[rule_id]["doctests"].sort(key=lambda x: x["result"], reverse=True)
        for doctest_index, doctest in enumerate(rules_dict[rule_id]["doctests"]):
            rules_dict[rule_id]["doctests"][doctest_index]["doctest_no"] = (
                doctest_index + 1
            )
    return rules_dict


def calculate_rules_doctest_stats(rules_dict: Dict[str, Dict]) -> Dict[str, Dict]:
    """
    Calculate aggregated passing/failing results of rules doctests
    """

    for rule_id in rules_dict:
        rule = rules_dict[rule_id]
        if rule["doctests"]:
            passing_doctests = len(
                [
                    doctest
                    for doctest in rule["doctests"]
                    if doctest["result"] == "True\n"
                ]
            )
            failing_doctests = len(
                [
                    doctest
                    for doctest in rule["doctests"]
                    if doctest["result"] == "False\n"
                ]
            )

            if not passing_doctests:
                passing_doctests = "None"
            if not failing_doctests:
                failing_doctests = "None"

            doctest_stats = [
                {
                    "title": "Pass",
                    "doctest_result": "True\n",
                    "total": passing_doctests,
                },
                {
                    "title": "Fail",
                    "doctest_result": "False\n",
                    "total": failing_doctests,
                },
            ]
            rules_dict[rule_id]["doctest_stats"] = doctest_stats

    return rules_dict


def enrich_rules(rules: List) -> Dict[str, Dict]:
    """
    Enriches a list of Rule objects with dependency information from
    rule children and parents
    """
    parsed_rules = [rule.parse() for rule in rules]
    enriched_rules = calculate_rules_depends_on(parsed_rules)
    enriched_rules_dict = {rule["id"]: rule for rule in enriched_rules}
    enriched_rules_dict = calculate_rules_breadcrumbs(enriched_rules_dict)
    enriched_rules_dict = enrich_rules_doctests_metadata(enriched_rules_dict)
    enriched_rules_dict = calculate_rules_doctest_stats(enriched_rules_dict)
    return enriched_rules_dict



import functools
import inspect
from typing import TYPE_CHECKING, Any, AnyStr, Dict, List, Optional

import regex as re

from ruleau.constants import OverrideLevel
from ruleau.context_manager import Context
from ruleau.doc_parsers import clean_source, comments, description, doctests, parameters
from ruleau.exceptions import (
    CannotOverrideException,
    ConditionalDependencyReusedException,
    RuleErrorException,
    RuleFunctionNameAttributeCollisionException,
    RuleIdIllegalCharacterException,
    RuleRequiresIdException,
    RuleRequiresNameException,
    RunIfAndRunIfNotSameRule,
    RunIfRuleHasRunIfOrNotException,
)
from ruleau.structures import ExecutionResult, RuleauDict

if TYPE_CHECKING:
    from ruleau.adapter import ApiAdapter
    from ruleau.process import Process
    from ruleau.types import Function


class Rule:
    def __init__(
        self,
        func: "Function",
        id_: AnyStr,
        name: AnyStr,
        depends_on: List["Rule"],
        override_level: OverrideLevel = OverrideLevel.ENABLED,
        run_if: Optional["Rule"] = None,
        run_if_not: Optional["Rule"] = None,
    ):
        """
        :param func: User defined rule
        :param name: User defined human readable name of the rule
        :param depends_on: Rule dependencies
        :param override_level: Override level
        :param run_if: Optional, Conditional rule
        :param run_if_not: Optional, Conditional rule
        """
        self.id = id_
        self.name = name
        # Set the user defined function
        self.func = func

        # Set exception debug information
        info = Rule._get_relevant_call_stack_info()
        self.line_number = info["lineno"]
        self.file_name = info["filename"]

        # Validate the rule, make sure the name is always set for a rule
        self._validate()

        # Set the override level for rule
        self.override_level = override_level
        # Set rule function name
        self.__name__ = func.__name__

        # Setup the rule dependencies
        self.depends_on = depends_on
        self.run_if = run_if
        self.run_if_not = run_if_not

        # Set default values for order & result
        self.order = None
        self.execution_result = None

        # Validate dependencies
        self._validate_duplicate_dependencies()
        self._validate_run_if_or_if_not()

        # This preserves the original Docstring on the decorated function
        # which allows DocTest to detect the function
        functools.update_wrapper(self, func)

    @property
    def dependencies(self):

        if self.run_if and self.run_if_not:
            return self.depends_on + [self.run_if, self.run_if_not]
        if self.run_if:
            return self.depends_on + [self.run_if]
        if self.run_if_not:
            return self.depends_on + [self.run_if_not]

        return self.depends_on

    def __str__(self) -> str:
        return f"{str(self.id)} - {str(self.name)} ({str(self.__name__)})"

    def __repr__(self):
        return str(self)

    def __call__(self, *args, **kwargs) -> bool:
        return self.func(*args, **kwargs)

    @staticmethod
    def _get_relevant_call_stack_info():
        found_rule_py = False
        result = {"lineno": -1, "filename": ""}
        stack = inspect.stack()
        for frame in stack:
            # The most helpful file to report is after rule.py and optionally,
            # decorators.py
            filename = frame.filename
            if filename.endswith("/rule.py") or filename.endswith("\\rule.py"):
                found_rule_py = True
                continue
            if found_rule_py and not filename.endswith("decorators.py"):
                result["lineno"] = frame.lineno
                result["filename"] = filename
                break

        return result

    def _validate(self):
        """
        Validator to check if top level rule has a human readable name
        and id
        :raises: TopLevelRuleRequiresNameException
        :raises: RuleRequiresIdException
        """
        # Validate if Rule name is present
        if not self.name or not isinstance(self.name, str):
            raise RuleRequiresNameException(rule=self)

        # Validate if Rule ID is present
        if not self.id or not isinstance(self.id, str):
            raise RuleRequiresIdException(rule=self)

        # Validate the Rule ID
        if not re.match(r"^([a-zA-Z0-9-_~]+)+$", self.id):
            raise RuleIdIllegalCharacterException(rule=self)

        # If an attribute with the same name as the function name
        try:
            getattr(self, self.func.__name__)
            raise RuleFunctionNameAttributeCollisionException(rule=self)
        except AttributeError:
            pass

    def _validate_duplicate_dependencies(self):
        """
        Validator to check if rule has duplicate dependencies
        in run_if and depends_on
        :raises: ConditionalDependencyReusedException
        """
        if (self.run_if and self.run_if_not) and (self.run_if == self.run_if_not):
            raise RunIfAndRunIfNotSameRule(self)

        # Validate if there are no duplicate dependencies
        for rule_depends in [self.run_if, self.run_if_not]:
            if rule_depends and rule_depends in self.depends_on:
                raise ConditionalDependencyReusedException(rule=self)

    def _validate_run_if_or_if_not(self):
        conditionals = Rule._get_conditionals(self)
        if conditionals:
            for cond in conditionals:
                if sub_cons := Rule._get_conditionals(cond):
                    raise RunIfRuleHasRunIfOrNotException(
                        rule=self, run_if_or_not=cond, sub_conditionals=sub_cons
                    )

    def _get_source(self) -> AnyStr:
        """Get the cleaned source code of the rule"""
        return clean_source(inspect.getsource(self.func))

    @staticmethod
    def _get_conditionals(current_rule):
        if current_rule is None:
            return []
        results = []
        if current_rule.run_if:
            results.append(current_rule.run_if)
        if current_rule.run_if_not:
            results.append(current_rule.run_if)
        return results

    @property
    def description(self) -> AnyStr:
        """Parse the description of rule"""
        return description(self.func.__doc__)

    def calculate_order_values(self, index=None):
        """
        This sets the `order` value for this rule
        and all of the dependencies below it.

        :param index: Used internally, do not set
        """
        if index is None:
            self.order = 0
        else:
            if self.order is not None:
                self.order = max(self.order, index + 1)
            else:
                self.order = index + 1
        for dependency in self.dependencies:
            dependency.calculate_order_values(self.order)

    def flatten_rule_objects(self, flat_rules=None) -> List["Rule"]:
        """Method to flatten rule objects, usually used for root rule"""
        if flat_rules is None:
            flat_rules = []
        flat_rules.append(self)
        for dependency in self.dependencies:
            dependency.flatten_rule_objects(flat_rules)
        return flat_rules

    def parse(self):
        """Collects and returns parsed rule data for documentation
        :return: Dictionary of rule data
        """
        return {
            "id": self.id,
            "name": self.name,
            "rule_type": type(self).__name__,
            "order": self.order,
            "override_level_name": self.override_level.name,
            "override_level": self.override_level.value,
            "source": self._get_source(),
            "comments": comments(self._get_source()),
            "docstring": self.func.__doc__,
            "description": self.description,
            "parameters": parameters(self.func.__doc__),
            "dependencies": self.sort_dependencies(),
            "depends_on": [],
            "doctests": doctests(self.func),
        }

    def sort_dependencies(self):
        """
        Sort the dependencies so that those with conditional rules come first.
        Returns a list of dicts suitable for display.
        """
        return sorted(
            [
                {
                    "id": dependent.id,
                    "run_if": dependent == self.run_if,
                    "run_if_not": dependent == self.run_if_not,
                }
                for dependent in self.dependencies
            ],
            key=lambda d: d["run_if"] or d["run_if_not"],
            reverse=True,
        )

    def execute(
        self,
        case_id: AnyStr,
        payload: Dict[AnyStr, Any],
        process: "Process",
        api_adapter: Optional["ApiAdapter"] = None,
    ) -> ExecutionResult:
        """
        Execute the rule and apply overrides
        :param case_id: Case ID
        :param payload: Case payload
        :param process: Process instance
        :param api_adapter: ApiAdapter instance
        :return: ExecutionResult of the rule
        """

        # Test the results of run_if rules
        run_if_result = True
        if self.run_if:
            run_if_result = self.run_if.execute(
                case_id=case_id,
                payload=payload,
                process=process,
                api_adapter=api_adapter,
            ).result

        if self.run_if_not:
            result = self.run_if_not.execute(
                case_id=case_id,
                payload=payload,
                process=process,
                api_adapter=api_adapter,
            ).result
            run_if_result = (not result) and run_if_result

        # if it fails then skip the execution of this rule
        if not run_if_result:
            self.skip_execution()
            self.remove_stale_data()
            return self.execution_result

        # If the rule is already executed return the previous result
        # unless that result was skipped
        if (
            self.execution_result
            and not self.execution_result.skipped
            and payload == self.execution_result.payload
        ):
            return self.execution_result

        # Prep the rule payload
        rule_payload = RuleauDict(payload)

        # Run this rule dependencies
        dependency_results = {
            dependency: dependency.execute(
                case_id=case_id,
                payload=payload,
                process=process,
                api_adapter=api_adapter,
            )
            for dependency in self.depends_on
        }
        context = Context(dependency_results)

        self.execution_result = ExecutionResult(
            self,
            rule_payload,
            None,
            dependency_results,
        )

        # Run this rule
        try:
            rule_result = self.__call__(context, rule_payload)
            self.execution_result.result = rule_result
        # If it fails mark it as failed
        except Exception as e:
            self.execution_result.failed = True
            raise RuleErrorException(rule=self, rule_exception=e, case_id=case_id)

        if api_adapter:
            # Apply overrides on the rule result
            self.execution_result = self.apply_override(
                case_id, process, self.execution_result, api_adapter
            )

        # Assign to internal object & return the rule result
        return self.execution_result

    def apply_override(
        self,
        case_id,
        process,
        execution_result: ExecutionResult,
        api_adapter: "ApiAdapter",
    ) -> ExecutionResult:
        """
        Check if there are any overrides to apply to this case. If so apply the
        override.
        :param case_id: Case ID
        :param execution_result: ExecutionResult of a rule
        :param api_adapter: ApiAdapter instance
        :return: ExecutionResult of the rule
        """
        # Get overrides for the rule in a case
        override = api_adapter.fetch_override(case_id, process.id, self.id)

        # Apply override to the executed rule result, if any
        # Overrides should only be applied to allowed rule and if they're present
        if override:
            # Throw an exception if the backend is trying to override a DISABLED rule
            if self.override_level == OverrideLevel.DISABLED:
                raise CannotOverrideException(case_id=case_id, rule=self)
            else:
                # Override the rule result and set the overridden flag
                execution_result.override = override["id"]
                execution_result.original_result = execution_result.result
                execution_result.result = override["applied"]
        return execution_result

    def skip_execution(self):
        """
        Recursively skip a rule and all its children. If a rule is skipped when
        it already has a result, that result is kept.
        """
        if self.execution_result is None:
            self.execution_result = ExecutionResult.skipped_result(self)
        for dependency in self.depends_on:
            dependency.skip_execution()

    def remove_stale_data(self):
        """
        Recursively removes stale data from nested Rules
        """

        self.execution_result.failed = False
        self.execution_result.original_result = None
        self.execution_result.override = None
        self.execution_result.payload = None
        self.execution_result.result = None
        self.execution_result.skipped = True

        for dependency in self.depends_on:
            dependency.remove_stale_data()



__version__ = "0.5.0"

from .adapter import ApiAdapter
from .aggregates import All, Any
from .constants import OverrideLevel
from .context_manager import mock_context
from .decorators import rule
from .execute import execute
from .structures import ExecutionResult



from collections import OrderedDict
from typing import TYPE_CHECKING, Any, AnyStr, Dict, Optional

from ruleau.exceptions import RuleErrorException
from ruleau.rule import Rule

if TYPE_CHECKING:
    from ruleau.adapter import ApiAdapter


class Process:
    """Class holding rule process"""

    def __init__(
        self,
        process_id: str,
        name: str,
        description: str,
        root_rule: Rule,
    ):
        self.id = process_id
        self.name = name
        self.description = description
        self.root_rule = root_rule

    @staticmethod
    def create_process_from_rule(rule: Rule) -> "Process":
        return Process(
            process_id=rule.id,
            name=rule.name,
            description=rule.description,
            root_rule=rule,
        )

    @property
    def rules(self):
        return list(OrderedDict.fromkeys(self.root_rule.flatten_rule_objects()))

    def parse(self):
        self.root_rule.calculate_order_values()

        return {
            "id": self.id,
            "name": self.name,
            "description": self.description,
            "root_rule": self.root_rule.id,
            "rules": [rule.parse() for rule in self.rules],
        }

    def execute(
        self,
        case_id: AnyStr,
        payload: Dict[AnyStr, Any],
        api_adapter: Optional["ApiAdapter"] = None,
    ):
        try:
            result = self.root_rule.execute(
                case_id=case_id,
                payload=payload,
                process=self,
                api_adapter=api_adapter,
            )
            return result
        # If any exception is hit during the process execution
        except RuleErrorException as e:
            raise e.rule_exception
        finally:
            # Always sync the rules current state
            if api_adapter:
                api_adapter.sync_results(self, case_id)



import logging
import os
from datetime import datetime, timedelta
from typing import Any, AnyStr, Dict, List, Optional, TypedDict
from urllib.parse import urljoin

import jwt
import requests

from ruleau.decorators import api_request
from ruleau.exceptions import (
    CaseAPIException,
    OrganisationalDataApiException,
    RuleAPIException,
    UiLayoutMetadataApiException,
)
from ruleau.process import Process
from ruleau.rule import Rule
from ruleau.structures import OrganisationalData, OrganisationalScheme, UiLayoutMetadata

logger = logging.getLogger(__name__)


class OrganisationalDataRow(TypedDict):
    key: str
    value: Any


class ReexecutionType(TypedDict):

    process_id: str
    case_id: str
    last_override_time: datetime


class ApiAdapter:
    base_url: AnyStr
    base_path: AnyStr
    username: Optional[AnyStr]
    password: Optional[AnyStr]
    organisational_data: Optional[List[OrganisationalData]]

    def __init__(
        self,
        base_url: AnyStr,
        username: Optional[AnyStr] = None,
        password: Optional[AnyStr] = None,
    ):
        """
        :param base_url: Base URL of the ruleau API
        :param username: (Optional) Users username
        :param password: (Optional) Users password
        """
        self.base_url = base_url
        self.base_path = "/api/v1/"

        self.username = os.getenv("RULEAU_USERNAME", username)
        self.password = os.getenv("RULEAU_PASSWORD", password)
        if not self.username or not self.password:
            raise ValueError("Username or Password not supplied")
        self.organisational_data = None

        self.access_token = None
        self.refresh_token = None
        self.access_token_expiry = None
        self.session = None
        self._fetch_jwt_token()

    def with_organisational_data(
        self, organisational_data: Optional[List[Dict]] = None
    ) -> "ApiAdapter":
        """
        Adds organisational_data to the ApiAdapter object.
        :param organisational_data: Organisation-specific data for display
        """
        self.organisational_data = organisational_data
        return self

    @api_request
    def sync_case(
        self,
        case_id: AnyStr,
        process_id: AnyStr,
        payload: Dict,
    ) -> Dict:
        """
        Synchronise case with API. Includes previously-saved organisational data which
        is cleared after sending.
        :param case_id: The ID of the case being executed
        :param process_id: The ID of the process
        :param payload: Case payload to execute on
        :return:
        """

        response = self.session.get(
            urljoin(
                self.base_url,
                f"{self.base_path}processes/{process_id}/cases/{case_id}",
            ),
        )

        if response.status_code == 200:
            response = self.session.patch(
                urljoin(
                    self.base_url,
                    f"{self.base_path}processes/{process_id}/cases/{case_id}",
                ),
                json={
                    "id": case_id,
                    "organisational_data": self.organisational_data,
                    "payload": payload,
                    "status": "OPEN",
                },
            )
            # Clear organisation data to prevent adding it where it doesn't belong.
            self.organisational_data = None
            if response.status_code != 200:
                raise CaseAPIException(
                    activity="update", case_id=case_id, response=response
                )

        elif response.status_code == 404:
            response = self.session.post(
                urljoin(self.base_url, f"{self.base_path}processes/{process_id}/cases"),
                headers={"Authorization": f"Bearer {self.access_token}"},
                json={
                    "id": case_id,
                    "organisational_data": self.organisational_data,
                    "payload": payload,
                    "process": process_id,
                    "status": "OPEN",
                },
            )
            # Clear organisation data to prevent adding it where it doesn't belong.
            self.organisational_data = None
            if response.status_code != 201:
                raise CaseAPIException(
                    activity="create", case_id=case_id, response=response
                )

        else:
            raise CaseAPIException(activity="check", case_id=case_id, response=response)

        return response.json()

    @api_request
    def publish_organisational_scheme(
        self, rule: Rule, organisational_scheme: List[OrganisationalScheme]
    ) -> List[Dict]:
        """
        Sets the organisational data fields that can be attached to a Process.
        :param rule: The rule used to identify the Process the fields relate to.
        :param organisational_scheme: The fields to set for the process.
        :return: The organisational schemes stored in the API.
        """
        process = Process.create_process_from_rule(rule)
        self.sync_process(process)

        url = urljoin(
            self.base_url,
            f"{self.base_path}processes/{process.id}/organisational_scheme",
        )

        response = self.session.post(
            url,
            json=organisational_scheme,
        )

        if response.status_code != 201:
            raise OrganisationalDataApiException(
                process_id=process.id, response=response
            )

        return response.json()

    @api_request
    def publish_ui_layout_metadata(
        self, rule: Rule, ui_layout_metadata: UiLayoutMetadata
    ) -> Dict:
        """
        Stores the order by which fields declared in the organisational scheme
        for a given Process should appear on both the Cases and Overrides screens.
        Any fields not specified in the layout metadata will appear after fields that
        are, ordered by the date they were created.
        :param rule: The rule used to identify the Process the fields relate to.
        :param ui_layout_metadata: Defines the order the fields should appear.
        :return: The metadata stored in the API.
        """
        process = Process.create_process_from_rule(rule)

        url = urljoin(
            self.base_url,
            f"{self.base_path}processes/{process.id}/ui_layout_metadata",
        )

        case_org_data_order = (
            [x["id"] for x in ui_layout_metadata["case_org_data_order"]]
            if "case_org_data_order" in ui_layout_metadata
            else []
        )

        override_org_data_order = (
            [x["id"] for x in ui_layout_metadata["override_org_data_order"]]
            if "override_org_data_order" in ui_layout_metadata
            else []
        )

        response = self.session.post(
            url,
            json={
                "case_org_data_order": case_org_data_order,
                "override_org_data_order": override_org_data_order,
            },
        )

        if response.status_code != 201:
            raise UiLayoutMetadataApiException(process_id=process.id, response=response)

        return response.json()

    @api_request
    def sync_process(self, process: Process):

        response = self.session.post(
            urljoin(self.base_url, f"{self.base_path}processes"),
            headers={"Authorization": f"Bearer {self.access_token}"},
            json=process.parse(),
        )

        if response.status_code != 201:
            raise RuleAPIException(
                activity="save", process_id=process.id, response=response
            )

        return response.json()

    @api_request
    def sync_results(
        self,
        process: "Process",
        case_id: AnyStr,
    ):
        payload = [
            {
                "rule": rule.id,
                "result": rule.execution_result.result,
                "payloads": rule.execution_result.payload.accessed
                if rule.execution_result.payload
                else None,
                "override": rule.execution_result.override,
                "original_result": rule.execution_result.original_result,
                "skipped": rule.execution_result.skipped,
            }
            for rule in process.rules
            if rule.execution_result
        ]
        response = self.session.post(
            urljoin(
                self.base_url,
                f"{self.base_path}processes/{process.id}/cases/" f"{case_id}/results",
            ),
            headers={"Authorization": f"Bearer {self.access_token}"},
            json=payload,
        )
        if response.status_code > 299:
            raise RuleAPIException(
                activity="store rule result for",
                process_id=process.id,
                response=response,
            )
        return None

    @api_request
    def fetch_override(
        self, case_id: AnyStr, process_id: AnyStr, rule_id: AnyStr
    ) -> Optional[Dict[AnyStr, Any]]:
        """
        Fetch rule overrides
        :param case_id: client ID that identifies a previously established case
        :param process_id: The ID of the process that the case is being run against
        :param rule_id: The ID of the Rule to fetch overrides for
        :return: a ruleau overrides Optional[Dict[AnyStr, Any]]
        """

        response = self.session.get(
            urljoin(
                self.base_url,
                f"{self.base_path}processes/{process_id}/"
                f"cases/{case_id}/overrides/search",
            ),
            headers={"Authorization": f"Bearer {self.access_token}"},
            params={"rule_id": rule_id},
        )
        if response.status_code != 200:
            return {}
        return response.json()

    def _fetch_jwt_token(self):
        """
        Fetches the JWT token for the username and password provided
        """
        login_response = requests.post(
            urljoin(self.base_url, f"{self.base_path}token/"),
            data={"username": self.username, "password": self.password},
        )
        if login_response.status_code != 200:
            raise requests.exceptions.RequestException(login_response.json())

        body = login_response.json()
        self.access_token = body["access"]
        self.refresh_token = body["refresh"]
        self._set_access_token_expiry(body)

    def _set_access_token_expiry(self, body):
        """
        Decodes the expiry time for the access token
        """
        access_payload = jwt.decode(body["access"], options={"verify_signature": False})
        self.access_token_expiry = datetime.fromtimestamp(int(access_payload["exp"]))

    def _refresh_access_token(self):
        """
        Gets a new access token using the refresh token
        """

        refresh_response = requests.post(
            urljoin(self.base_url, f"{self.base_path}token/refresh/"),
            data={
                "refresh": self.refresh_token,
            },
        )
        if refresh_response.status_code != 200:
            raise requests.exceptions.RequestException(refresh_response.json())

        body = refresh_response.json()
        self.access_token = body["access"]
        self._set_access_token_expiry(body)

    def _check_access_token_is_active(self):
        """
        Checks to see if the access token is close to expiring (within 5 seconds)
        or has expired
        """
        time_now = datetime.now()
        if time_now > (self.access_token_expiry - timedelta(seconds=5)):
            self._refresh_access_token()

    def _create_requests_session(self):
        self.session = requests.session()
        self.session.headers.update({"Authorization": f"Bearer {self.access_token}"})

    @api_request
    def fetch_cases_for_reexecution(self) -> [ReexecutionType]:

        response = requests.get(
            urljoin(
                self.base_url,
                f"{self.base_path}cases/reexecution",
            ),
            headers={"Authorization": f"Bearer {self.access_token}"},
        )
        if response.status_code != 200:
            return []
        return response.json()



from examples.kitchen_sink.rules import will_lend
from ruleau import execute

if __name__ == "__main__":
    payload = {
        "data": {
            "fico_score": 150,
            "ccjs": [],
            "kyc": "low",
            "number_of_children": 1,
            "capital": 10_000,
            "ccjs_required": True,
        }
    }
    result = execute(will_lend, payload, case_id="abc")

    print(f"result: {result.result}")



from typing import AnyStr, Optional

from ruleau.constants import OverrideLevel
from ruleau.rule import Rule
from ruleau.structures import ExecutionResult


class Any(Rule):
    """Aggregator to implement OR operation
    Returns truthy, if any one of the rule result is truthy
    """

    def __init__(
        self,
        rule_id: AnyStr,
        name: AnyStr,
        *rules: "Rule",
        description: AnyStr = None,
        override_level: OverrideLevel = OverrideLevel.ENABLED,
        run_if: Optional[Rule] = None,
        run_if_not: Optional[Rule] = None,
    ):
        def any_aggregator(context: ExecutionResult, _):
            if all([result.skipped for result in context.dependent_results]):
                self.skip_execution()
                self.remove_stale_data()
                return None
            return any(result.result for result in context.dependent_results)

        any_aggregator.__doc__ = description if description else Any.__doc__
        super().__init__(
            any_aggregator,
            rule_id,
            name,
            list(rules),
            override_level,
            run_if,
            run_if_not,
        )


class All(Rule):
    """Aggregator to implement AND operation
    Returns truthy, if all of the rule results are truthy
    """

    def __init__(
        self,
        rule_id: AnyStr,
        name: AnyStr,
        *rules: "Rule",
        description: AnyStr = None,
        override_level: OverrideLevel = OverrideLevel.ENABLED,
        run_if: Optional[Rule] = None,
        run_if_not: Optional[Rule] = None,
    ):
        def all_aggregator(context: ExecutionResult, _):
            if all([result.skipped for result in context.dependent_results]):
                self.skip_execution()
                self.remove_stale_data()
                return None
            return all(
                result.result
                for result in context.dependent_results
                if not result.skipped
            )

        all_aggregator.__doc__ = description if description else All.__doc__
        super().__init__(
            all_aggregator,
            rule_id,
            name,
            list(rules),
            override_level,
            run_if,
            run_if_not,
        )


class AllFail(Rule):
    """Aggregator to implement NONE/Not Any/All Fail/None shall pass operation
    Returns true, if all of the rule results are false
    """

    def __init__(
        self,
        rule_id: AnyStr,
        name: AnyStr,
        *rules: "Rule",
        description: AnyStr = None,
        override_level: OverrideLevel = OverrideLevel.ENABLED,
        run_if: Optional[Rule] = None,
        run_if_not: Optional[Rule] = None,
    ):
        def all_fail_aggregator(context: ExecutionResult, _):
            if all([result.skipped for result in context.dependent_results]):
                self.skip_execution()
                self.remove_stale_data()
                return None
            return all(
                result.result is False
                for result in context.dependent_results
                if not result.skipped
            )

        all_fail_aggregator.__doc__ = description if description else AllFail.__doc__
        super().__init__(
            all_fail_aggregator,
            rule_id,
            name,
            list(rules),
            override_level,
            run_if,
            run_if_not,
        )



from typing import TYPE_CHECKING, AnyStr, Dict, List, Optional

if TYPE_CHECKING:
    from ruleau.rule import Rule
    from ruleau.structures import RuleauDict


class ExecutionResult:
    def __init__(
        self,
        executed_rule: "Rule",
        payload: "RuleauDict",
        result,
        dependent_results: Dict["Rule", "ExecutionResult"],
        override: AnyStr = None,
        original_result: Optional[bool] = None,
        skipped: bool = False,
        failed: bool = False,
    ):
        self.rule = executed_rule
        self.payload = payload
        self.result = result
        self.override = override
        self.original_result = original_result
        self.dependent_results = dependent_results
        self.skipped = skipped
        self.failed = failed

    @property
    def rule_results(self) -> List["ExecutionResult"]:
        """Returns flat list of ExecutionResults for this rule and
        all descendants of this rule"""
        return self._flatten_rule_results()

    def _flatten_rule_results(self, flat_rule_results=None) -> List["ExecutionResult"]:
        if flat_rule_results is None:
            flat_rule_results = []
        if self not in flat_rule_results:
            flat_rule_results.append(self)
        for execution_result in self.dependent_results.values():
            execution_result._flatten_rule_results(flat_rule_results)
        return flat_rule_results

    @staticmethod
    def skipped_result(rule: "Rule") -> "ExecutionResult":
        """Creates an ExecutionResult for a skipped rule"""
        return ExecutionResult(rule, None, None, {}, skipped=True)



from ruleau.structures.execution_result import ExecutionResult
from ruleau.structures.org_data_models import (
    OrganisationalData,
    OrganisationalScheme,
    OrganisationalSchemeID,
    UiLayoutMetadata,
)
from ruleau.structures.ruleau_dict import RuleauDict



from logging import getLogger
from typing import AnyStr, List, Optional

import requests

from ruleau.constants import OverrideLevel
from ruleau.rule import Rule

logger = getLogger(__name__)
logger.propagate = True


def api_request(func):
    def _api_request(*args, **kwargs) -> Optional[dict]:
        try:
            if args:
                args[0]._check_access_token_is_active()
                args[0]._create_requests_session()
            return func(*args, **kwargs)
        except Exception as e:
            logger.error(f"Exception occurred: {e}")
            raise e

    return _api_request


def rule(
    rule_id: AnyStr,
    name: AnyStr,
    depends_on: Optional[List[Rule]] = None,
    override_level: OverrideLevel = OverrideLevel.ENABLED,
    run_if: Optional[Rule] = None,
    run_if_not: Optional[Rule] = None,
):
    """Decorator to encapsulate a function into a rule"""
    depends_on = depends_on or []

    def _rule(func) -> Rule:
        return Rule(func, rule_id, name, depends_on, override_level, run_if, run_if_not)

    return _rule



from examples.kitchen_sink.rules import will_lend
from ruleau import ApiAdapter, execute

if __name__ == "__main__":
    # Add API adapter to runner
    result = execute(
        will_lend,
        {
            "data": {
                "fico_score": 150,
                "ccjs": [],
                "kyc": "low",
                "number_of_children": 1,
                "capital": 10_000,
                "ccjs_required": True,
            }
        },
        api_adapter=ApiAdapter(
            base_url="http://127.0.0.1:8000"
        ).with_organisational_data([{"key": "Data Name", "value": None}]),
        case_id="abc",
    )



from typing import List, Literal, TypedDict, Union

SchemeTypes = Literal["string", "integer", "float", "date"]


class OrganisationalData(TypedDict):
    key: str
    value: Union[str, int, float]


class OrganisationalSchemeID(TypedDict):
    id: str


class OrganisationalScheme(OrganisationalSchemeID):
    display_name: str
    display_default: bool
    type: SchemeTypes


class UiLayoutMetadata(TypedDict):
    case_org_data_order: List[OrganisationalSchemeID]
    override_org_data_order: List[OrganisationalSchemeID]



from typing import Callable, Optional, Union

from ruleau.structures import ExecutionResult, RuleauDict

Function = Callable[
    [Optional[ExecutionResult], Optional[RuleauDict]], Union[None, bool]
]



from rule_based_block_rules import account_status
from ruleau import execute

if __name__ == "__main__":
    execution_result = execute(
        account_status,
        {
            "loc_record": {
                "loc_number": 12345,
                "current_dnp": 0,
                "current_cons_full_pmt": 4,
                "expected_payments_passed": 1,
                "expected_payments_future": 15,
                "total_pending_payment": 1234.51,
                "total_balance": 2345.21,
                "principal_balance": 1995.85,
                "plaid_active_flag": True,
                "draw": [
                    {
                        "draw_date": "2019-01-01T00:00:00.000Z",
                        "amount": 5000.01,
                        "principal_post_draw": 8123.50,
                    },
                    {
                        "draw_date": "2019-06-01T00:00:00.000Z",
                        "amount": 4000.02,
                        "principal_post_draw": 5123.50,
                    },
                ],
                "missed_pmt": [
                    {
                        "missed_pmt_cluster": 1,
                        "payment_date": "2021-01-01T00:00:00.000Z",
                    },
                    {
                        "missed_pmt_cluster": 1,
                        "payment_date": "2020-12-26T00:00:00.000Z",
                    },
                    {
                        "missed_pmt_cluster": 2,
                        "payment_date": "2020-11-01T00:00:00.000Z",
                    },
                ],
                "modification": [
                    {
                        "applied": "2020-01-01T00:00:00.000Z",
                        "suspended": "2020-02-01T00:00:00.000Z",
                        "mod_cluster": 1,
                        "reduction_perc": 50,
                        "instalment_amount": 200.05,
                    },
                    {
                        "applied": "2020-02-01T00:00:00.000Z",
                        "suspended": "2020-02-05T00:00:00.000Z",
                        "mod_cluster": 1,
                        "reduction_perc": 100,
                        "instalment_amount": 0,
                    },
                    {
                        "applied": "2020-03-01T00:00:00.000Z",
                        "suspended": "2020-01-01T00:00:00.000Z",
                        "mod_cluster": 2,
                        "reduction_perc": 75,
                        "instalment_amount": 100.03,
                    },
                ],
            }
        },
    )

    for rule_result in execution_result.rule_results:
        print(
            "{0} - {1} - {2}".format(
                rule_result.rule.id.ljust(20),
                str(rule_result.result).ljust(5),
                rule_result.rule.name,
            )
        )



import collections
import json
import re
import sys
from doctest import DocTestFinder
from textwrap import dedent
from typing import TYPE_CHECKING, AnyStr, List, Optional

from ruleau.context_manager import mock_context
from ruleau.exceptions import IncorrectKwargsForDoctests

if TYPE_CHECKING:
    from ruleau.types import Function

PARAM_REGEX_PATTERN = ":(?P<name>[\*\s\w]+): (?P<description>.*?)\Z"  # noqa: W605


def clean_source(source: AnyStr) -> AnyStr:
    """Clean the source code.

    Remove all the lines until we have passed the decorators, function header
    and docstring.

    :param source: The source code to clean.
    """
    # Remove common indent
    source = dedent(source)

    # Get lines from source
    lines = source.split("\n")

    # Get line function header starts on. All function sources must have a header
    header_index = next(
        idx for idx, line in enumerate(lines) if line.startswith("def")
    )  # pragma: no cover

    # Remove anything above header from lines (decorators) and header it self
    lines = lines[header_index + 1 :]  # noqa: E203

    # Remove the docstring
    doc_string_start_idx = get_start_of_docstring(lines)

    # If we found a doc string
    if doc_string_start_idx is not None:
        doc_string_end_idx = get_end_of_docstring(lines, doc_string_start_idx)
        # Remove the doc string lines
        lines = lines[doc_string_end_idx + 1 :]  # noqa: E203

    clean_source_ = dedent("\n".join(lines))
    return clean_source_


def comments(source: AnyStr) -> AnyStr:
    """Returns single line comments from provided source

    :param source: clean source (output of clean_source)
    :returns: returns single line comments from source code
    """
    output_lines = [
        line for line in source.splitlines() if line.strip().startswith("#")
    ]
    return "\n".join(output_lines)


def context_as_dict(context):
    items = context.dependent_results.dependencies.items()
    return {
        "dependent_results": {
            rule_object: execution_result.result
            for rule_object, execution_result in items
        }
    }


def description(docstring: AnyStr) -> AnyStr:
    """Takes a rule docstring, and returns just the lines
    representing a description

    :param docstring: Docstring for a function
    """
    result = trim(docstring or "")
    lines = result.split("\n")
    end_of_description_idx = next(
        (
            idx
            for idx, line in enumerate(lines)
            if any(
                [
                    # Line is empty i.e. new line after description
                    line == "" or line.isspace(),
                    # Line is the start of a doc string parameter
                    re.match(PARAM_REGEX_PATTERN, line),
                    # Line is the start of a doc test
                    line.startswith(">>>"),
                ]
            )
        ),
        # If no end to the description then it ends with the string
        len(lines),
    )

    description_ = "\n".join(lines[:end_of_description_idx])
    # Remove final new line character if present
    return remove_suffix(description_, "\n")


def doctests(func: "Function") -> List[dict]:
    """Extract the doctest from a rule
    :param func: Function wrapped in a Rule instance
    :return: List of doctest examples
    """

    doctests = []
    finder = DocTestFinder(verbose=False, recurse=False)
    #  For each test
    for tests in finder.find(func):
        # Collect the source code up to the first result is expected
        source = ""
        for example in tests.examples:
            source = source + example.source

            # Once a result is expected run the source and extract the provided
            # values.
            if example.want.strip() != "":
                if source_is_example(example.source, func.__name__):
                    context, payload = get_doctest_rule_args(func, source, tests.globs)
                    context_json = (
                        json.dumps(context_as_dict(context)) if context else None
                    )
                    doctests.append(
                        {
                            "test": source,
                            "result": example.want,
                            "context": context_json,
                            "payload": json.dumps(payload),
                        }
                    )
                source = ""

    return doctests


def get_doctest_rule_args(func: "Function", source: str, globs=None):
    """Given a rule function and the source, extract the arguments used when
    calling the rule in the given source.

    :param func: The rule function
    :param source: The source that calls the function
    :param globs: Globals for the rule execution
    """

    context = None
    payload = None

    #  Mock for rule function
    def mock_func(*args, **kwargs):
        nonlocal context
        nonlocal payload
        nonlocal func
        #  We dont return early as the arguments for the final call of the
        # function should be used

        payload_kwarg = "payload"

        if args and not kwargs:
            context, payload = args

        elif kwargs and not args:
            context_kwarg = "context" if "context" in kwargs.keys() else "_"
            target_kwargs = [context_kwarg, payload_kwarg]
            if all(kwarg in kwargs.keys() for kwarg in target_kwargs):
                context = kwargs[context_kwarg]
                payload = kwargs[payload_kwarg]
            else:
                raise IncorrectKwargsForDoctests(func)
        else:
            context = args[0]
            if payload_kwarg in kwargs.keys():
                payload = kwargs["payload"]
            else:
                raise IncorrectKwargsForDoctests(func)

    # Execute the function with the mocked rule
    mocks = {func.__name__: mock_func, "mock_context": mock_context}
    globs = globs or {}
    globs = {**globs, **mocks}
    exec(source, globs)

    # Return the context and payload used in the final call of the rule function
    return context, payload


def get_end_of_docstring(source_lines: List[AnyStr], start_idx=0) -> int:
    """Get the index of line on which the doc string ends.

    :param source_lines: A list of the lines of the source of the function
    without the header or decorators
    :param start_idx: The line on which the docstring starts

    :raise: Exception if no end of docstring found
    """
    for line_number, line in enumerate(source_lines[start_idx:]):  # pragma: no cover
        # A docstring cannot be started and not ended

        count = line.count('"""')  # Number of occurances of """ on line

        # If we are on the starting line (same as doc_string_start_idx) we
        # need to find 2 """ for the doc string to be closed.
        if (line_number == 0 and count == 2) or (line_number != 0 and count >= 1):
            return start_idx + line_number

    raise ValueError("No end of docstring found")


def get_index_of_closing_bracket(source):
    """Returns index of closing bracket
    :param: source starting with opening bracket
    :returns: returns the index of the closing bracket that matches the opening
    bracket
    """
    # Create a deque to use it as a stack.
    queue = collections.deque()

    for i in range(len(source)):
        # Pop a starting bracket
        # for every closing bracket
        if source[i] == ")":
            queue.popleft()

        # Push all starting brackets
        elif source[i] == "(":
            queue.append(source[i])

        # If deque becomes empty
        if not queue:
            return i

    return None


def get_start_of_docstring(source_lines: List[AnyStr]) -> Optional[int]:
    """Get the index of line on which the doc string starts.

    :param source_lines: A list of the lines of the source of the function
    without the header or decorators
    """
    # Loop to find the start of the doc string
    for idx, line in enumerate(source_lines):

        # Skip any empty lines
        if line == "" or line.isspace():  # pragma: no cover
            # This is covered in the tests but no picked up by coverage
            continue

        # Find the start of the doc string
        if line.strip().startswith('"""'):
            return idx

        # If we find any non empty line that is not starting with """ then
        # there is no doc string for this function
        else:
            return None


def parameters(docstring: AnyStr):
    """Parse list of parameters out of the docstring

    :param docstring: The docstring of the rule
    :returns: [{"name": ..., "description": ...}, ...]
    """
    param_regex = re.compile(PARAM_REGEX_PATTERN)

    params = []
    if docstring:
        lines = docstring.split("\n")
        for line in lines:
            match = param_regex.findall(line.strip())
            if match:
                param_name, param_description = match[0]
                params.append({"name": param_name, "value": param_description})

    return params


def remove_prefix(source, prefix):
    """Removes prefix from a string source.
    :param source: String to remove prefix from
    :param prefix: Prefix to remove from the given string
    :return: a string with prefix removed
    """
    if source.startswith(prefix):
        return source[len(prefix) :]  # noqa: E203
    else:
        return source


def remove_suffix(source, suffix):
    """Removes suffix from a source string.
    :param source: String to remove suffix from
    :param suffix: Suffix to remove from the given string
    :return: a string with suffix removed
    """
    if source.endswith(suffix):
        return source[: -len(suffix)]
    else:
        return source


def source_is_example(source, function_name):

    # Remove function name
    source = source.strip()
    if not source.startswith(function_name):
        return False
    source = remove_prefix(source, function_name)
    source = source.strip()

    # Check after function name we find (
    if not source.startswith("("):
        return False

    # Find closing bracket of example
    closing_bracket_index = get_index_of_closing_bracket(source)
    if closing_bracket_index is None:
        return False

    #  If more happens after the call to the function
    if not source[closing_bracket_index:].strip() == ")":
        return False

    return True


def trim(docstring):
    """trim function implementation from PEP-257
    https://www.python.org/dev/peps/pep-0257/#id18

    :param docstring: Docstring for a function
    """
    if not docstring:
        return ""
    # Convert tabs to spaces (following the normal Python rules)
    # and split into a list of lines:
    lines = docstring.expandtabs().splitlines()
    # Determine minimum indentation (first line doesn't count):
    indent = sys.maxsize
    for line in lines[1:]:
        stripped = line.lstrip()
        if stripped:
            indent = min(indent, len(line) - len(stripped))
    # Remove indentation (first line is special):
    trimmed = [lines[0].strip()]
    if indent < sys.maxsize:  # pragma: no cover
        for line in lines[1:]:
            trimmed.append(line[indent:].rstrip())
    # Strip off trailing and leading blank lines:
    while trimmed and not trimmed[-1]:
        trimmed.pop()
    while trimmed and not trimmed[0]:
        trimmed.pop(0)

    if "\n" in docstring:
        trimmed.append("")
    # Return a single string:
    return "\n".join(trimmed)



import inspect
import re
from os import path
from typing import List, Optional


def rule_exception_message(msg: str, rules_list: List) -> str:
    """Writes an extended debug message for rules-based exceptions."""
    msgs_list = []
    msg = "" if msg is None else msg
    for a_rule in rules_list:
        if a_rule is not None and hasattr(a_rule, "id"):
            msgs_list.append(
                f"Rule id {a_rule.id} on line {a_rule.line_number} in file "
                f"{a_rule.file_name}"
            )
        else:
            msgs_list.append("Unknown rule")

    return msg + " Rule detail: " + ", ".join(msgs_list)


def get_relevant_locals(f_locals: dict) -> str:
    """
    Grabs only those dictionary items where the key contains one of the
    relevant_keywords
    """
    relevant_keywords = ["case_id", "payload", "process_id"]
    relevant_locals = {}
    for k, v in f_locals.items():
        if any([keyword in k for keyword in relevant_keywords]):
            relevant_locals[k] = v

    return ", ".join([f"{k}: {v}" for k, v in relevant_locals.items()])


def find_first_non_exception_frame(call_stack: List) -> Optional[inspect.FrameInfo]:
    """
    Returns the first non '*exception*.py' frame in the call stack.
    This should be the interesting one and is needed when we can't get the info via the
    constructor.
    """
    result = None
    for frame in call_stack:
        if "exception" not in path.basename(frame.filename):
            result = frame
            break

    return result


def get_message_from_call_stack(call_stack: List, msg: str) -> str:
    useful_frame = find_first_non_exception_frame(call_stack)
    exception_info = get_relevant_locals(f_locals=useful_frame.frame.f_locals)
    return (
        f"APIException in file: '{useful_frame.filename}', "
        f"function: {useful_frame.function}, "
        f"line: '{useful_frame.lineno}', with exception info: '{exception_info}', "
        f"message: '{msg}'"
    )


def get_rule_detail_text(rule):
    return (
        f"The rule with ID {rule.id} named {rule.name} on {rule.line_number} in "
        f"{rule.file_name}"
    )


class RuleException(Exception):
    """Generic exceptions from which all Rule-Exceptions can inherit."""

    def __init__(self, msg=None, rule_1=None, rule_2=None):
        self.message = rule_exception_message(msg=msg, rules_list=[rule_1, rule_2])


class RuleRequiresNameException(RuleException):
    """Exception raised if a rule doesn't have a human readable name"""

    def __init__(self, rule):
        if hasattr(rule, "id") and rule.id:
            self.message = (
                f"Rule on {rule.line_number} in {rule.file_name} with ID {rule.id} is "
                f"missing a name."
            )
        else:
            self.message = (
                f"There is a rule on {rule.line_number} in {rule.file_name} that is "
                f"missing a name."
            )


class RuleRequiresIdException(RuleException):
    """Exception raised if rule doesn't have an id"""

    def __init__(self, rule):
        if hasattr(rule, "name") and rule.name:
            self.message = (
                f"Rule on {rule.line_number} in {rule.file_name} with name {rule.name}"
                f" has no ID."
            )
        else:
            self.message = (
                f"There is a rule on {rule.line_number} in {rule.file_name} that has "
                f"no ID."
            )


class MethodNotAllowedException(RuleException):
    """Exception raised if a forbidden RuleauDict method is called"""

    def __init__(self, method, rule_dict):
        method_text = "delete" if method == "__delitem__" else "set"
        self.message = (
            f"A rule on {rule_dict.line_number} in {rule_dict.file_name} tried to "
            f"{method_text} a value in the dict which is not allowed."
        )


class CaseIdRequiredException(RuleException):
    """Exception raised if a json path for case identifier is not found"""

    def __init__(self, rule):
        super().__init__(
            msg=f"A case id for rule: {rule.id} was not provided which required when "
            f"posting data to the API.",
            rule_1=rule,
        )


class CannotOverrideException(Exception):
    """Exception raised if a API tries to override a rule marked as NO_OVERRIDE"""

    def __init__(self, case_id, rule):
        self.message = (
            f"Case ID {case_id} is set to override but {rule.name} on "
            f"{rule.line_number} in {rule.file_name} prohibits overrides."
        )


class DuplicateRuleNameException(RuleException):
    """Exception raised if more than 1 rule has same name"""

    def __init__(self, rule, conflict):
        # We can assume that the rule has a valid ID, given that Rule._validate()
        # will have been called from the constructor and raised an exception if it
        # wasn't.
        self.message = (
            f"The name {rule.name} on {rule.line_number} in {rule.file_name} with "
            f"id {rule.id} is in conflict with rule {conflict.id} on "
            f"{conflict.line_number} in {conflict.file_name}"
        )


class RuleIdIllegalCharacterException(RuleException):
    """Exception raise if Rule name contains illegal characters"""

    def __init__(self, rule):
        rule_detail = get_rule_detail_text(rule)
        illegal_chars = re.findall(r"[^a-zA-Z0-9-_~]", rule.id)
        self.message = (
            f"{rule_detail} has illegal characters in the ID: {illegal_chars}"
        )


class RuleFunctionNameAttributeCollisionException(RuleException):
    """If the provided rule name collides with a rule attribute"""

    def __init__(self, rule):
        self.message = (
            f"{get_rule_detail_text(rule)} in conflict with a function "
            f"with the same name."
        )


class DuplicateRuleIdException(RuleException):
    """Exception raised if more than 1 rule has same ID"""

    def __init__(self, rule, conflict):
        self.message = (
            f"{get_rule_detail_text(rule)} is in conflict with rule {conflict.id} on "
            f"{conflict.line_number} in {conflict.file_name}"
        )


class APIException(Exception):
    """Generic exception for API request failure"""

    def __init__(self, msg):
        self.message = get_message_from_call_stack(call_stack=inspect.stack(), msg=msg)


class CaseAPIException(APIException):
    def __init__(self, activity, case_id, response):
        msg = f"Failed to {activity} case {case_id} due to {response.text}"
        super().__init__(msg=msg)


class RuleAPIException(APIException):
    def __init__(self, activity, process_id, response):
        msg = (
            f"Failed to {activity} rules in process {process_id} due to "
            f"{response.text}"
        )
        super().__init__(msg=msg)


class OrganisationalDataApiException(APIException):
    def __init__(self, process_id, response):
        msg = (
            f"Failed to add the Organisational Scheme for process {process_id} due to "
            f"{response.text}"
        )
        super().__init__(msg=msg)


class UiLayoutMetadataApiException(APIException):
    def __init__(self, process_id, response):
        msg = (
            f"Failed to add the UI Layout Metadata for process {process_id} due to "
            f"{response.text}"
        )
        super().__init__(msg=msg)


class ConditionalDependencyReusedException(RuleException):
    """Exception raised if the rule has duplicate dependencies"""

    def __init__(self, rule):
        run_if_line = rule.run_if.line_number if rule.run_if is not None else "None"
        run_if_file = rule.run_if.file_name if rule.run_if is not None else "None"
        run_if_not_line = (
            rule.run_if_not.line_number if rule.run_if_not is not None else "None"
        )
        run_if_not_file = (
            rule.run_if_not.file_name if rule.run_if_not is not None else "None"
        )
        super().__init__(
            msg=(
                f"{get_rule_detail_text(rule)} "
                f" has {run_if_line}/{run_if_not_line} in "
                f"{run_if_file}/{run_if_not_file} linked as a run_if/run_if_not as a "
                f"dependency."
            ),
            rule_1=rule,
            rule_2=rule.run_if,
        )


class RuleErrorException(Exception):
    """Exception raised when a rule raises an exception"""

    def __init__(self, rule, rule_exception, case_id):
        self.rule_exception = rule_exception
        self.rule_id = rule.id
        self.rule_name = rule.name
        self.message = (
            f"{get_rule_detail_text(rule)} "
            f"has thrown an exception ({rule_exception}) with {case_id}"
        )


class RunIfRuleHasRunIfOrNotException(RuleException):
    """
    Exception raise if a run_if or run_if_not dependency of a rule also has a run_if or
    run_if_not.
    """

    def __init__(self, rule, run_if_or_not, sub_conditionals):
        self.message = (
            f"{get_rule_detail_text(rule)} has run_if/run_if_not: {run_if_or_not.id} "
            f"at {run_if_or_not.line_number} in {run_if_or_not.file_name} which is "
            f"linked to other(s) conditionals "
            f"{', '.join([sc.id for sc in sub_conditionals])} which is not a valid "
            f"state."
        )


class RunIfAndRunIfNotSameRule(RuleException):
    """The required keynames context and payload have not been provided"""

    def __init__(self, rule):
        self.rule_name = rule.name
        self.rule_id = rule.id
        self.message = (
            f"{get_rule_detail_text(rule)} has RUN_IF and RUN_IF_NOT which is linked "
            f"to the same rule which is {rule.run_if}/{rule.run_if_not}"
        )


class IncorrectKwargsForDoctests(Exception):
    """The required keynames context and payload have not been provided"""

    def __init__(self, rule):
        self.rule_id = rule.id
        self.rule_name = rule.name
        self.message = (
            f"{get_rule_detail_text(rule)} "
            f"has not been defined with the correct keyword argument names. Please "
            f"change the rule argument names to context and payload."
        )
        super().__init__(self.message)


class ImportException(Exception):
    """
    Class of exception that occurs when attempting to import
    a Ruleau ruleset for documentation
    """

    def __init__(self, module_file: str, top_level_rule: str, message: str):
        self.module_file = module_file
        self.top_level_rule = top_level_rule
        self.message = message
        super().__init__(message)


class CouldNotImportRulesToDocument(ImportException):
    """Ruleau Docs could not load the provided module and top level rule"""

    def __init__(
        self,
        module_file: str,
        top_level_rule: str,
        inner_exception: ModuleNotFoundError,
    ):
        self.inner_exception = inner_exception

        message = (
            f"Could not import the file '{module_file}' "
            f"with the top level rule '{top_level_rule}' to be documented.\n"
            f"Was unable to import a module the file depends "
            f"on called '{inner_exception.name}'"
        )

        super().__init__(module_file, top_level_rule, message)


class IncorrectTypeProvidedForImport(ImportException):
    """
    Ruleau Docs could not load the provided top level rule as it is not a Rule instance
    """

    def __init__(self, module_file: str, top_level_rule: str):

        message = (
            f"Could not import the file '{module_file}' "
            f"with the top level rule '{top_level_rule}' to be documented.\n"
            f"The provided top level rule isn't a Rule type, it must be an "
            f"instance of a rule (for example an All or Any rule)"
        )

        super().__init__(module_file, top_level_rule, message)



import argparse
import importlib
import json
import os
import sys
from argparse import Namespace
from os import mkdir, path
from typing import AnyStr, List, Optional

from jinja2 import Template
from jinja2.filters import FILTERS, environmentfilter

from ruleau.documentation_enrichment_functions import enrich_rules
from ruleau.exceptions import (
    CouldNotImportRulesToDocument,
    IncorrectTypeProvidedForImport,
)
from ruleau.rule import Rule


@environmentfilter
def json_print(_, value):
    """Pretty print json filter for jinja2"""
    value = value if isinstance(value, dict) or value is None else json.loads(value)
    return json.dumps(
        value,
        sort_keys=True,
        indent=4,
        separators=(",", ": "),
    )


@environmentfilter
def filter_runif_dependencies(_, dependencies):
    return [dep for dep in dependencies if dep["run_if"]]


@environmentfilter
def filter_non_runif_dependencies(_, dependencies):
    return [dep for dep in dependencies if not dep["run_if"]]


#  Add the json_print filter to jinja2 filters
FILTERS["json_print"] = json_print
FILTERS["filter_runif"] = filter_runif_dependencies
FILTERS["filter_non_runif"] = filter_non_runif_dependencies


def order_rules(rules: List["Rule"]) -> List["Rule"]:
    """Order rules so top level rules appear first"""
    is_root = {rule: True for rule in rules}

    # If any rule is dependent on a rule then it is not the root
    for rule in rules:
        for dependent in rules:
            if dependent != rule and dependent in rule.dependencies:
                is_root[dependent] = False

    roots = [rule for rule, is_root_rule in is_root.items() if is_root_rule]
    others = [rule for rule, is_root_rule in is_root.items() if not is_root_rule]
    return roots + others


def generate_documentation(rules: List["Rule"]) -> AnyStr:
    """Returns a HTML string, documenting the passed in rule and its dependents

    :param rules: List of rules to generate document of
    """
    with open(
        path.join(path.dirname(path.realpath(__file__)), "html", "documentation.html")
    ) as f:
        doc_template = Template(f.read())

    logo_path = path.join(path.dirname(path.realpath(__file__)))

    logo_path = path.join(logo_path, "html", "images", "RuleauLogo_Black.png")

    rules = order_rules(rules)
    enriched_rules_dict = enrich_rules(rules)

    #  for rule  in enriched_rules_dict:
    #  colors = {
    #  "light": "75FF96",
    #  "primary": "12C170",
    #  "dark": "17764A",
    #  }

    return doc_template.render(rules=enriched_rules_dict, logo_path=logo_path)


def get_dependencies(rule: "Rule", rule_list: Optional[List] = None):
    """Recursively adds the rule dependencies to a list
    :param rule: The Rule to add dependencies for.
    :param: rule_list: An optional list of dependencies that
     are already known to exist.
    :return: rule_list: A list of Rule dependencies.
    """
    if rule_list is None:
        rule_list = []
    if rule not in rule_list:
        rule_list.append(rule)
    for dependency in rule.dependencies:
        get_dependencies(dependency, rule_list)
    return rule_list


def render_doc_for_module(module_file, top_level_rule) -> AnyStr:
    """Render the documentation from a rule module
    :param module_file:
    :param top_level_rule: The top level rule that is dependent on all other rules.
    :return:
    """
    # Need the path of the rules
    rule_path = path.split(path.abspath(module_file))[0]

    # Module name is different to the module path
    module_name = path.splitext(module_file)[0]
    # Formatting for the module registering
    # Formatting for a Windows Path
    module_name = module_name.replace("\\", ".")
    # Formatting for a Nix path
    module_name = module_name.replace("/", ".")
    # Remove any trailing "." characters (e.g: from when './module.py' is passed in)
    module_name = module_name.lstrip(".")

    sys.path.append(rule_path)

    spec = importlib.util.spec_from_file_location(module_name, module_file)
    config = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = config

    # Append the current working directory for any package imports
    # in the referenced script
    cwd = os.getcwd()
    if cwd not in sys.path:
        sys.path.append(cwd)

    try:
        spec.loader.exec_module(config)
    except ModuleNotFoundError as ex:
        raise CouldNotImportRulesToDocument(module_file, top_level_rule, ex)

    rule_instance = getattr(config, top_level_rule)

    if not isinstance(rule_instance, Rule):
        raise IncorrectTypeProvidedForImport(module_file, top_level_rule)

    # Pick rules to document
    rules = get_dependencies(rule_instance)
    return generate_documentation(rules)


def generate_and_save_to_file(input_files, output_dir):
    """Generate the documentation and save to a file in provided output dir
    :param input_files: Rule modules and top level function names provided by
     user via CLI
    :param output_dir: Output directory of generated HTML file
    """

    generated_docs = []

    if len(input_files) == 0:
        raise Exception("No file(s) supplied to generate documentation for.")

    if not path.exists(output_dir):
        mkdir(output_dir)

    for target_file in input_files:
        if target_file.split("::") == [target_file]:
            raise ValueError(
                "Please provide a name of top level function. "
                f"Required format is {target_file}::function_name."
            )
        top_level_rule = target_file.split("::")[1]
        target_file = target_file.split("::")[0]
        if not path.exists(target_file):
            raise ValueError(f"{target_file} does not exist")

        documentation = render_doc_for_module(target_file, top_level_rule)

        output_filename = path.join(
            output_dir, "{}.html".format(path.basename(target_file).split(".")[0])
        )

        with open(output_filename, "w") as f:
            f.write(documentation)
        generated_docs.append(output_filename)

    print(
        f"{len(generated_docs)} doc{'s' if len(generated_docs) > 1 else ''} generated."
    )
    print("\n".join(generated_docs))


def get_arguments(args) -> Namespace:
    """Parses arguments for the ruleau-docs command
    :param args:
    :return:
    """
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--output-dir",
        dest="output_dir",
        default="./",
        help="Directory for generated documents",
    )
    parser.add_argument("files", nargs="*")
    return parser.parse_args(args)


def main() -> int:  # pragma: no cover
    """Console script for deft document generation.
    USAGE:
    ```bash
    $ ruleau-docs [--output-dir=<argument>] filename ...
    ```
    """
    args = get_arguments(sys.argv[1:])
    generate_and_save_to_file(args.files, args.output_dir)
    return 0


if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



import itertools
import re
from pathlib import Path

import setuptools

with open("ruleau/__init__.py", "r") as f:
    version = re.search(
        r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', f.read(), re.MULTILINE
    ).group(1)

with open(Path(__file__).parent / "README.md", encoding="utf-8") as f:
    long_description = f.read()

with open("./requirements.txt", "r") as r:
    requirements = r.readlines()

setuptools.setup(
    name="ruleau",
    version=version,
    author="Unai Ltd",
    author_email="pypi@unai.com",
    description="A python rules engine",
    license="BSD 3-Clause",
    long_description=long_description,
    long_description_content_type="text/markdown",
    tests_require=["pytest"],
    url="https://gitlab.com/unai-ltd/unai-decision/ruleau-core",
    packages=setuptools.find_packages(exclude=("tests",)),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6.8",
    entry_points={"console_scripts": ["ruleau-docs=ruleau.docs:main"]},
    package_data={"": ["html/documentation.html"]},
    include_package_data=True,
    install_requires=requirements,
)



from lending_rules import (
    ccjs_check_required,
    fico_score_greater_than_threshold,
    has_no_ccjs,
    has_sufficient_capital,
    kyc_risk_greater_than_threshold,
)

from ruleau import All, ApiAdapter, execute, rule


@rule("rul-101A", "Causes skipped")
def causes_skip(_, __):
    return False


@rule("rul_101", "Skipped Rule", run_if=causes_skip)
def skipped_rule(_, __):
    return False


@rule("rul-102A", "Causes not skipped")
def causes_no_skip(_, __):
    return True


@rule("rul_102", "Not Skipped Rule", run_if=causes_no_skip)
def not_skipped_rule(_, __):
    return False


@rule("rul200-c", "Skipped tree bottom")
def skipped_tree_bottom(_, __):
    return False


@rule(
    "rul200-a",
    "Skipped tree 1  skipped",
    depends_on=[skipped_tree_bottom],
    run_if=causes_skip,
)
def skipped_tree_1_a(_, __):
    return False


@rule(
    "rul200-b",
    "Skipped tree 1 not skipped",
    depends_on=[skipped_tree_bottom],
    run_if=causes_no_skip,
)
def skipped_tree_1_b(_, __):
    return False


@rule("rul200", "Skipped tree 1 Top", depends_on=[skipped_tree_1_a, skipped_tree_1_b])
def skipped_tree_1(_, __):
    return False


conditionals = All(
    "ID",
    "name",
    skipped_rule,
    causes_no_skip,
    skipped_tree_1,
)

will_lend = All(
    "WL_01",
    "Will Lend",
    kyc_risk_greater_than_threshold,
    fico_score_greater_than_threshold,
    has_no_ccjs,
    has_sufficient_capital,
    conditionals,
)

if __name__ == "__main__":
    result = execute(
        will_lend,
        {
            "data": {
                "fico_score": 150,
                "ccjs": [],
                "kyc": "low",
                "number_of_children": 1,
                "capital": 10_000,
                "ccjs_required": True,
            }
        },
        api_adapter=ApiAdapter(base_url="http://127.0.0.1:8000"),
        case_id="132",
    )
    print("Result", result.result)



import logging
from typing import TYPE_CHECKING, Any, AnyStr, Dict, List, Optional

from jsonpath_ng import parse

from ruleau.adapter import ApiAdapter
from ruleau.exceptions import (
    CaseIdRequiredException,
    DuplicateRuleIdException,
    DuplicateRuleNameException,
)
from ruleau.process import Process
from ruleau.rule import Rule

if TYPE_CHECKING:
    from ruleau.structures import ExecutionResult

logger = logging.getLogger(__name__)


def validate_no_duplicate_rule_names(rules: List[Rule]) -> None:
    """Returns True if there are no duplicate Rule Names are used
    A name can only be re-used if the same rule is included multiple times
    """
    rules_dict = {}
    for rule in rules:
        if rule.name not in rules_dict:
            rules_dict[rule.name] = rule
        else:
            if rule != rules_dict[rule.name]:
                raise DuplicateRuleNameException(rule, rules_dict[rule.name])


def validate_no_duplicate_rule_ids(rules: List[Rule]) -> None:
    """Returns True if there are no duplicate Rule IDs used
    An ID can only be re-used if the same rule is included multiple times
    """
    rules_dict = {}
    for rule in rules:
        if rule.id not in rules_dict:
            rules_dict[rule.id] = rule
        else:
            if rule != rules_dict[rule.id]:
                raise DuplicateRuleIdException(rule, rules_dict[rule.id])


def get_case_id_from_json_path(
    payload: Dict[AnyStr, Any],
    case_id_jsonpath: AnyStr,
) -> Optional[str]:
    """
    Return case_id value from json payload

    :param payload Dict[AnyStr, Any]: Raise
    :param case_id_jsonpath AnyStr: The json path to the case_id in the payload
    :rtype Optional[str]: The case_id from the payload (or None if no path provided)
    :raises ValueError: If provided JSON path not found
    :raises ValueError: If the case_id at the path is blank (empty string)
    """
    # If jsonpath is None, then there is no case_id
    if case_id_jsonpath is None:
        return None

    # Otherwise find the case id in the json
    case_id_results = parse(case_id_jsonpath).find(payload)
    if not case_id_results:
        raise ValueError("Case ID not found in payload")

    case_id = str(case_id_results[0].value)
    if not case_id:
        raise ValueError("Case ID not found")

    return case_id


def execute(
    executable_rule: Rule,
    payload: Dict[AnyStr, Any],
    case_id_jsonpath: AnyStr = None,
    case_id: Optional[AnyStr] = None,
    api_adapter: Optional[ApiAdapter] = None,
) -> "ExecutionResult":
    """
    Executes the provided rule, following dependencies and
    passing in results accordingly
    """

    if api_adapter:
        case_id = (
            case_id
            if case_id is not None
            else get_case_id_from_json_path(payload, case_id_jsonpath)
        )
        if not case_id:
            raise CaseIdRequiredException(rule=executable_rule)

    # Validate unique rule name
    flattened_rules_as_objects = executable_rule.flatten_rule_objects()
    validate_no_duplicate_rule_names(flattened_rules_as_objects)

    # Validate unique rule ids
    validate_no_duplicate_rule_ids(flattened_rules_as_objects)

    # If API adapter was passed sync the case
    executable_rule.calculate_order_values()
    process = Process.create_process_from_rule(executable_rule)

    if api_adapter:

        # Sync the process rules
        api_adapter.sync_process(process)

        # Sync the case
        api_adapter.sync_case(case_id=case_id, process_id=process.id, payload=payload)

    # Trigger the rule execution, from the top level rule
    return process.execute(
        case_id=case_id,
        payload=payload,
        api_adapter=api_adapter,
    )



import json
from typing import TYPE_CHECKING, AnyStr, Dict

from ruleau.structures import ExecutionResult

if TYPE_CHECKING:
    from ruleau.rule import Rule


class Context:
    """
    The context is passed in to the execution of a rule. It stores the results
    of the dependencies of the rule.
    """

    class DependentResults:
        """
        Used primarily to support iterating through all the results using
        `for result in context.dependent_results` in a rule
        """

        def __init__(self, execution_results: Dict["Rule", "ExecutionResult"]):
            self.dependencies = execution_results

        def _get_execution_result(self, rule):
            try:
                return self.dependencies[rule]
            except KeyError as e:
                # Although it would be nice to share the function name, that's
                # not necessarily available to us (e.g. if the rule is created)
                # using a factory function such as the agreggators
                raise AttributeError(
                    f"Result for rule '{rule}' not available, as it was not "
                    f"declared as a dependency'. "
                    f"Update the `depends_on` attribute of this rule to include"
                    f"the required object (currently depending on rule ids "
                    f"{json.dumps(list(rule.id for rule in self.dependencies.keys()))}"
                    f" )",
                    e,
                )

        def __iter__(self):
            # Iterate through the dependencies
            for dep in self.dependencies:
                yield self._get_execution_result(dep)

    def __init__(self, execution_results: Dict["Rule", "ExecutionResult"]):
        self.execution_results = execution_results
        self.dependent_results = self.DependentResults(execution_results)

    def _get_execution_result(self, rule):
        return self.dependent_results._get_execution_result(rule)

    def get_result(self, rule):
        return self._get_execution_result(rule).result


def mock_context(rule_results: Dict[str, bool]):
    """
    Given an dictionary containing the rule function name
    as the key and the mocked result as the value, creates
    an ExecutionResult.

    :param rule_results: A dictionary containing the mock
    results for the depedent rules

    :return: ExecutionResult containing mocked results
    """

    class MockDependentResults(Context.DependentResults):
        def _get_execution_result(self, rule):
            return self.dependencies[rule.__name__]

    class MockContext(Context):
        def __init__(self, execution_results):
            self.execution_results = execution_results
            self.dependent_results = MockDependentResults(execution_results)

    return MockContext(
        {
            key: ExecutionResult(None, {}, value, {})
            for key, value in rule_results.items()
        }
    )



from datetime import datetime, timedelta

from ruleau import All, rule


@rule(rule_id="MP-001-B1", name="Days Not Paid")
def days_not_paid(_, payload):
    """
    Block if the latest payment is now late (i.e. Days not paid is greater than zero).

    >>> days_not_paid(None, {"loc_record": {"current_dnp": 1}})
    False
    >>> days_not_paid(None, {"loc_record": {"current_dnp": 0}})
    True
    """
    # Fail if days not paid is greater than zero.
    return payload["loc_record"]["current_dnp"] == 0


@rule(rule_id="MP-001-U1", name="Two Consecutive Payments Made")
def two_consecutive_payments_made(_, payload):
    """
    Unblock if two consecutive, non-reduced payments been made.
    :Override Guidance: Testing out override guidance param.

    >>> two_consecutive_payments_made(None, {"loc_record": {"current_cons_full_pmt": 1, "expected_payments_passed": 10}})
    False
    >>> two_consecutive_payments_made(None, {"loc_record": {"current_cons_full_pmt": 2, "expected_payments_passed": 10}})
    True
    >>> two_consecutive_payments_made(None, {"loc_record": {"current_cons_full_pmt": 0, "expected_payments_passed": 1}})
    False
    >>> two_consecutive_payments_made(None, {"loc_record": {"current_cons_full_pmt": 1, "expected_payments_passed": 1}})
    True
    """
    # Default the required number of consecutive payments to two.
    consecutive_payment_required = 2

    # If the account is new (and so can not yet meet the required number of consecutive payments) then lower the number required.
    if payload["loc_record"]["expected_payments_passed"] < consecutive_payment_required:
        consecutive_payment_required = payload["loc_record"]["expected_payments_passed"]

    # Fail if the current number of consecutive full payments is less than the required number of payments.
    return (
        payload["loc_record"]["current_cons_full_pmt"] >= consecutive_payment_required
    )


@rule(rule_id="MP-002-B1", name="Two Missed Payments in 90 Days")
def two_missed_payments(_, payload):
    """
    Block if there are missed two payments within the last 90 days.

    :Override Guidance: Test guidance


    >>> two_missed_payments(None, {"loc_record": {"missed_pmt": [
    ...            {
    ...                "missed_pmt_cluster": 1,
    ...                "payment_date": "2021-02-01T09:00:00.000Z"
    ...            },
    ...            {
    ...                "missed_pmt_cluster": 1,
    ...                "payment_date": "2021-01-01T12:00:00.000Z"
    ...            },
    ...            {
    ...                "missed_pmt_cluster": 2,
    ...                "payment_date": "2020-11-01T00:00:00.000Z"
    ...            }
    ...        ]}})
    True
    """

    # Identify all the missed payments within the last 90 days
    missed_payments_list = []
    for missed_pmt in payload["loc_record"]["missed_pmt"]:
        if datetime.strptime(missed_pmt["payment_date"], "%Y-%m-%dT%H:%M:%S.%fZ") >= (
            datetime.today() - timedelta(days=90)
        ):
            missed_payments_list.append(missed_pmt)

    # Fail if the number of missed payments identified is 2 or more.
    return len(missed_payments_list) < 2


@rule(rule_id="MP-002-U1", name="Three Consecutive Payments Made")
def three_consecutive_payments_made(_, payload):
    """
    Unblock if three consecutive, non-reduced payments have been made.

    >>> three_consecutive_payments_made(None, {"loc_record": {"current_cons_full_pmt": 2, "expected_payments_passed": 10}})
    False
    >>> three_consecutive_payments_made(None, {"loc_record": {"current_cons_full_pmt": 3, "expected_payments_passed": 10}})
    True
    >>> three_consecutive_payments_made(None, {"loc_record": {"current_cons_full_pmt": 0, "expected_payments_passed": 1}})
    False
    >>> three_consecutive_payments_made(None, {"loc_record": {"current_cons_full_pmt": 2, "expected_payments_passed": 2}})
    True
    """
    # Default the required number of consecutive payments to three.
    consecutive_payment_required = 3

    # If the account is new (and so can not yet meet the required number of consecutive payments) then lower the number required.
    if payload["loc_record"]["expected_payments_passed"] < consecutive_payment_required:
        consecutive_payment_required = payload["loc_record"]["expected_payments_passed"]

    # Fail if the current number of consecutive full payments is less than the required number of payments.
    return (
        payload["loc_record"]["current_cons_full_pmt"] >= consecutive_payment_required
    )


@rule(rule_id="MP-003-B1", name="Three Missed Payments in 90 Days")
def three_missed_payments(_, payload):
    """
    Block if there are three missed payments within the last 90 days.

    >>> three_missed_payments(None, {"loc_record": {"missed_pmt": [
    ...            {
    ...                "missed_pmt_cluster": 1,
    ...                "payment_date": "2021-08-01T09:00:00.000Z"
    ...            },
    ...            {
    ...                "missed_pmt_cluster": 1,
    ...                "payment_date": "2021-07-25T12:00:00.000Z"
    ...            },
    ...            {
    ...                "missed_pmt_cluster": 1,
    ...                "payment_date": "2021-07-18T00:00:00.000Z"
    ...            }
    ...        ]}})
    False
    >>> three_missed_payments(None, {"loc_record": {"missed_pmt": [
    ...            {
    ...                "missed_pmt_cluster": 1,
    ...                "payment_date": "2021-08-01T09:00:00.000Z"
    ...            },
    ...            {
    ...                "missed_pmt_cluster": 1,
    ...                "payment_date": "2021-07-01T12:00:00.000Z"
    ...            },
    ...            {
    ...                "missed_pmt_cluster": 2,
    ...                "payment_date": "2020-02-01T00:00:00.000Z"
    ...            }
    ...        ]}})
    True
    """

    # Identify all missed payments within the last 90 days.
    missed_payments_list = []
    for missed_pmt in payload["loc_record"]["missed_pmt"]:
        if datetime.strptime(missed_pmt["payment_date"], "%Y-%m-%dT%H:%M:%S.%fZ") >= (
            datetime.today() - timedelta(days=90)
        ):
            missed_payments_list.append(missed_pmt)

    # Fail if the number of missed payments identified is 3 or more.
    return len(missed_payments_list) < 3


@rule(rule_id="MP-003-U1", name="Four Consecutive Payments Made")
def four_consecutive_payments_made(_, payload):
    """
    Unblock if four consecutive, non-reduced payments have been made.

    >>> four_consecutive_payments_made(None, {"loc_record": {"current_cons_full_pmt": 3, "expected_payments_passed": 10}})
    False
    >>> four_consecutive_payments_made(None, {"loc_record": {"current_cons_full_pmt": 4, "expected_payments_passed": 10}})
    True
    >>> four_consecutive_payments_made(None, {"loc_record": {"current_cons_full_pmt": 2, "expected_payments_passed": 3}})
    False
    >>> four_consecutive_payments_made(None, {"loc_record": {"current_cons_full_pmt": 3, "expected_payments_passed": 3}})
    True
    """
    # Default the required number of consecutive payments to four.
    consecutive_payment_required = 4

    # If the account is new (and so can not yet meet the required number of consecutive payments) then lower the number required.
    if payload["loc_record"]["expected_payments_passed"] < consecutive_payment_required:
        consecutive_payment_required = payload["loc_record"]["expected_payments_passed"]

    # Fail if the current number of consecutive full payments is less than the required number of payments.
    return (
        payload["loc_record"]["current_cons_full_pmt"] >= consecutive_payment_required
    )


@rule(rule_id="MP-004-B1", name="More Than Three Missed Payments in 90 Days")
def more_than_three_missed_payments(_, payload):
    """
    Block if there are more than three missed payments within the last 90 days.


    >>> more_than_three_missed_payments(None, {"loc_record": {"missed_pmt": [
    ...            {
    ...                "missed_pmt_cluster": 1,
    ...                "payment_date": "2021-08-01T09:00:00.000Z"
    ...            },
    ...            {
    ...                "missed_pmt_cluster": 1,
    ...                "payment_date": "2021-07-25T12:00:00.000Z"
    ...            },
    ...            {
    ...                "missed_pmt_cluster": 2,
    ...                "payment_date": "2021-07-07T08:00:00.456Z"
    ...            },
    ...            {
    ...                "missed_pmt_cluster": 2,
    ...                "payment_date": "2021-06-30T15:45:15.000Z"
    ...            },
    ...            {
    ...                "missed_pmt_cluster": 3,
    ...                "payment_date": "2021-02-01T18:00:12.123Z"
    ...            }
    ...        ]}})
    False
    >>> more_than_three_missed_payments(None, {"loc_record": {"missed_pmt": [
    ...            {
    ...                "missed_pmt_cluster": 1,
    ...                "payment_date": "2021-08-08T09:00:00.000Z"
    ...            },
    ...            {
    ...                "missed_pmt_cluster": 1,
    ...                "payment_date": "2021-08-01T12:00:00.000Z"
    ...            },
    ...            {
    ...                "missed_pmt_cluster": 1,
    ...                "payment_date": "2021-07-25T15:00:00.000Z"
    ...            },
    ...            {
    ...                "missed_pmt_cluster": 2,
    ...                "payment_date": "2020-11-01T00:00:00.000Z"
    ...            }
    ...        ]}})
    True
    """

    # Identify all missed payments within the last 90 days.
    missed_payments_list = []
    for missed_pmt in payload["loc_record"]["missed_pmt"]:
        if datetime.strptime(missed_pmt["payment_date"], "%Y-%m-%dT%H:%M:%S.%fZ") >= (
            datetime.today() - timedelta(days=90)
        ):
            missed_payments_list.append(missed_pmt)

    # Fail if the number of missed payments identified is greater three.
    return len(missed_payments_list) < 4


@rule(rule_id="MP-004-U1", name="Six Consecutive Payments Made")
def six_consecutive_payments_made(_, payload):
    """
    Unblock if six consecutive, non-reduced payments have been made.

    >>> six_consecutive_payments_made(None, {"loc_record": {"current_cons_full_pmt": 6, "expected_payments_passed": 10}})
    True
    >>> six_consecutive_payments_made(None, {"loc_record": {"current_cons_full_pmt": 5, "expected_payments_passed": 10}})
    False
    >>> six_consecutive_payments_made(None, {"loc_record": {"current_cons_full_pmt": 5, "expected_payments_passed": 5}})
    True
    >>> six_consecutive_payments_made(None, {"loc_record": {"current_cons_full_pmt": 4, "expected_payments_passed": 5}})
    False
    """
    # Default the required number of consecutive payments to three.
    consecutive_payment_required = 6

    # If the account is new (and so can not yet meet the required number of consecutive payments) then lower the number required.
    if payload["loc_record"]["expected_payments_passed"] < consecutive_payment_required:
        consecutive_payment_required = payload["loc_record"]["expected_payments_passed"]

    # Fail if the current number of consecutive full payments is less than the required number of payments.
    return (
        payload["loc_record"]["current_cons_full_pmt"] >= consecutive_payment_required
    )


@rule(rule_id="PR-001-B1", name="Two Payment Reductions in 180 Days")
def two_payment_reductions_in_180_days(_, payload):
    """
    Block if there are two or more payment reductions within the last 180 days.

    >>> two_payment_reductions_in_180_days(None, {"loc_record": {
    ...         "modification": [
    ...             {
    ...                 "applied": "2021-07-01T00:00:00.000Z",
    ...                 "reduction_perc": 50
    ...             },
    ...             {
    ...                 "applied": "2021-06-01T00:00:00.000Z",
    ...                 "reduction_perc": 75
    ...             },
    ...         ]
    ...     }})
    True
    >>> two_payment_reductions_in_180_days(None, {"loc_record": {
    ...         "modification": [
    ...             {
    ...                 "applied": "2021-07-01T00:00:00.000Z",
    ...                 "reduction_perc": 50
    ...             },
    ...             {
    ...                 "applied": "2021-01-01T00:00:00.000Z",
    ...                 "reduction_perc": 75
    ...             },
    ...         ]
    ...     }})
    False
    """

    # Identify all the payment reductions applied within the last 180 days.
    payment_reductions_list = [
        x
        for x in payload["loc_record"]["modification"]
        if datetime.strptime(x["applied"], "%Y-%m-%dT%H:%M:%S.%fZ")
        >= (datetime.today() - timedelta(days=180))
        if x["reduction_perc"] > 0
    ]

    # Fail if 2 or more reductions were identified.
    return len(payment_reductions_list) >= 2


@rule(rule_id="PR-001-U1", name="50% of Principal Paid")
def fifty_percent_principal_paid(_, payload):
    """
    Unblock if 50% of the principal balance has been paid following the most recent payment.

    >>> fifty_percent_principal_paid(None, {"loc_record": {
    ...         "principal_balance": 5000,
    ...         "draw": [
    ...             {
    ...                 "draw_date": "2021-08-01T00:00:00.000Z",
    ...                 "amount": 3000.00,
    ...                 "principal_post_draw": 12000.00
    ...             },
    ...             {
    ...                 "draw_date": "2019-07-01T00:00:00.000Z",
    ...                 "amount": 2000.00,
    ...                 "principal_post_draw": 15000.00
    ...             }
    ...         ]
    ...     }})
    True
    >>> fifty_percent_principal_paid(None, {"loc_record": {
    ...         "principal_balance": 8000,
    ...         "draw": [
    ...             {
    ...                 "draw_date": "2021-08-01T00:00:00.000Z",
    ...                 "amount": 3000.00,
    ...                 "principal_post_draw": 12000.00
    ...             },
    ...             {
    ...                 "draw_date": "2019-07-01T00:00:00.000Z",
    ...                 "amount": 2000.00,
    ...                 "principal_post_draw": 15000.00
    ...             }
    ...         ]
    ...     }})
    False
    """
    # Calculate the required balance threshold (i.e. 50% of the actual balance following the latest draw).
    balance_threshold = 0
    if len(payload["loc_record"]["draw"]) > 0:
        balance_threshold = (
            payload["loc_record"]["draw"][0]["principal_post_draw"] * 0.5
        )

    # Fail if balance required is greater than the account principal balance.
    return payload["loc_record"]["principal_balance"] <= balance_threshold


@rule(rule_id="PR-002-B1", name="Two Payment Reductions in 90 Days")
def two_payment_reductions_in_90_days(_, payload):
    """
    Block if there are two or more payment reductions within the last 90 days.

    >>> two_payment_reductions_in_90_days(None, {"loc_record": {
    ...         "modification": [
    ...             {
    ...                 "applied": "2021-08-01T00:00:00.000Z",
    ...                 "reduction_perc": 50
    ...             },
    ...             {
    ...                 "applied": "2021-07-25T00:00:00.000Z",
    ...                 "reduction_perc": 75
    ...             },
    ...         ]
    ...     }})
    True
    >>> two_payment_reductions_in_90_days(None, {"loc_record": {
    ...         "modification": [
    ...             {
    ...                 "applied": "2021-07-01T00:00:00.000Z",
    ...                 "reduction_perc": 50
    ...             },
    ...             {
    ...                 "applied": "2021-01-01T00:00:00.000Z",
    ...                 "reduction_perc": 75
    ...             },
    ...         ]
    ...     }})
    False

    """

    # Identify all the payment reductions applied within the last 90 days.
    payment_reductions_list = [
        x
        for x in payload["loc_record"]["modification"]
        if datetime.strptime(x["applied"], "%Y-%m-%dT%H:%M:%S.%fZ")
        >= (datetime.today() - timedelta(days=90))
        if x["reduction_perc"] > 0
    ]

    # Fail if 2 or more reductions were identified.
    return len(payment_reductions_list) >= 2


@rule(rule_id="PR-002-U1", name="75% of Principal Paid")
def seventy_five_percent_principal_paid(_, payload):
    """
    Unblock if 75% of the principal balance has been paid following the most recent payment.

    >>> seventy_five_percent_principal_paid(None, {"loc_record": {
    ...         "principal_balance": 2500,
    ...         "draw": [
    ...             {
    ...                 "draw_date": "2021-08-01T00:00:00.000Z",
    ...                 "amount": 3000.00,
    ...                 "principal_post_draw": 12000.00
    ...             },
    ...             {
    ...                 "draw_date": "2019-07-01T00:00:00.000Z",
    ...                 "amount": 2000.00,
    ...                 "principal_post_draw": 15000.00
    ...             }
    ...         ]
    ...     }})
    True
    >>> seventy_five_percent_principal_paid(None, {"loc_record": {
    ...         "principal_balance": 8000,
    ...         "draw": [
    ...             {
    ...                 "draw_date": "2021-08-01T00:00:00.000Z",
    ...                 "amount": 3000.00,
    ...                 "principal_post_draw": 12000.00
    ...             },
    ...             {
    ...                 "draw_date": "2019-07-01T00:00:00.000Z",
    ...                 "amount": 2000.00,
    ...                 "principal_post_draw": 15000.00
    ...             }
    ...         ]
    ...     }})
    False
    """
    # Calculate the required balance threshold (i.e. 25% of the actual balance following the latest draw).
    balance_threshold = 0
    if len(payload["loc_record"]["draw"]) > 0:
        balance_threshold = (
            payload["loc_record"]["draw"][0]["principal_post_draw"] * 0.25
        )

    # Fail if balance required is greater than the account principal balance.
    return payload["loc_record"]["principal_balance"] <= balance_threshold


@rule(rule_id="PR-003-B1", name="Two Consecutive Payment Reductions")
def two_consecutive_payment_reductions(_, payload):
    """
    Block if there are two 2 consecutive payment reductions.

    >>> two_consecutive_payment_reductions(None, {"loc_record": {
    ...         "modification": [
    ...             {
    ...                 "mod_cluster": 1,
    ...                 "reduction_perc": 50,
    ...             },
    ...             {
    ...                 "mod_cluster": 2,
    ...                 "reduction_perc": 75,
    ...             },
    ...             {
    ...                 "applied": "2021-06-01T00:00:00.000Z",
    ...                 "mod_cluster": 3,
    ...                 "reduction_perc": 75,
    ...             }
    ...         ]
    ...     }})
    True
    >>> two_consecutive_payment_reductions(None, {"loc_record": {
    ...         "modification": [
    ...             {
    ...                 "mod_cluster": 1,
    ...                 "reduction_perc": 50,
    ...             },
    ...             {
    ...                 "mod_cluster": 1,
    ...                 "reduction_perc": 75,
    ...             },
    ...             {
    ...                 "mod_cluster": 2,
    ...                 "reduction_perc": 75,
    ...             }
    ...         ]
    ...     }})
    False
    """

    # Create a list of payment modifications grouped by their consecutive clustering
    clusters = {}
    for mod in payload["loc_record"]["modification"]:
        if mod["reduction_perc"] > 0:
            if str(mod["mod_cluster"]) not in clusters:
                clusters[str(mod["mod_cluster"])] = []
            clusters[str(mod["mod_cluster"])].append(mod)

    # Check if any of the clusters contain more than 1 payment modifications
    consecutive_reductions_found = False
    for c in clusters.keys():
        if len(clusters[c]) > 1:
            consecutive_reductions_found = True

    # Fail if any consecutive reductions were found
    return consecutive_reductions_found is not True


@rule(rule_id="PR-003-U2", name="No Reductions Since Latest Draw")
def no_reductions_since_latest_draw(_, payload):
    """
    Unblock if there have been payment reduction has taken place since the last time a draw was made on the account.

    >>> no_reductions_since_latest_draw(None, {"loc_record": {
    ...         "draw": [
    ...             {
    ...                 "draw_date": "2021-08-01T00:00:00.000Z",
    ...                 "amount": 5000.00,
    ...                 "principal_post_draw": 12000.00
    ...             },
    ...             {
    ...                 "draw_date": "2019-07-01T00:00:00.000Z",
    ...                 "amount": 2000.00,
    ...                 "principal_post_draw": 17000.00
    ...             }
    ...         ],
    ...         "modification": [
    ...             {
    ...                 "applied": "2021-07-01T00:00:00.000Z",
    ...                 "suspended": "2021-07-21T00:00:00.000Z",
    ...                 "mod_cluster": 1,
    ...                 "reduction_perc": 50,
    ...                 "instalment_amount": 200.05
    ...             },
    ...             {
    ...                 "applied": "2021-06-01T00:00:00.000Z",
    ...                 "suspended": "2021-06-14T00:00:00.000Z",
    ...                 "mod_cluster": 2,
    ...                 "reduction_perc": 75,
    ...                 "instalment_amount": 125.00
    ...             }
    ...         ]
    ...     }})
    True
    >>> no_reductions_since_latest_draw(None, {"loc_record": {
    ...        "draw": [
    ...            {
    ...                "draw_date": "2021-08-01T00:00:00.000Z",
    ...                "amount": 5000.00,
    ...                "principal_post_draw": 12000.00
    ...            },
    ...            {
    ...                "draw_date": "2019-07-01T00:00:00.000Z",
    ...                "amount": 2000.00,
    ...                "principal_post_draw": 17000.00
    ...            }
    ...        ],
    ...        "modification": [
    ...            {
    ...                "applied": "2021-08-8T00:00:00.000Z",
    ...                "suspended": "2021-08-25T00:00:00.000Z",
    ...                "mod_cluster": 1,
    ...                "reduction_perc": 50,
    ...                "instalment_amount": 200.05
    ...            },
    ...            {
    ...                "applied": "2021-06-01T00:00:00.000Z",
    ...                "suspended": "2021-06-14T00:00:00.000Z",
    ...                "mod_cluster": 2,
    ...                "reduction_perc": 75,
    ...                "instalment_amount": 125.00
    ...            },
    ...        ]
    ...    }})
    False
    """

    # Get the date of the latest draw. If there are no draws then use a default 'early date'.
    latest_draw_date = "1900-01-01T00:00:00Z"
    if len(payload["loc_record"]["draw"]) > 0:
        latest_draw_date = payload["loc_record"]["draw"][0]["draw_date"]

    # Identify any payment reductions that were applied after the latest draw date.
    payment_reductions_list = [
        x
        for x in payload["loc_record"]["modification"]
        if x["applied"] > latest_draw_date
        if x["reduction_perc"] > 0
    ]

    # Fail if any payment reductions were identified.
    return len(payment_reductions_list) == 0


@rule(rule_id="PR-004-B1", name="On Payment Holiday")
def on_payment_holiday(_, payload):
    """
    Block if the account is currently subject to an agreed payment holiday.

    >>> on_payment_holiday(None, {"loc_record": {
    ...         "modification": [
    ...             {
    ...                 "applied": "2021-07-01T00:00:00.000Z",
    ...                 "suspended": "2021-12-01T00:00:00.000Z",
    ...                 "instalment_amount": 0
    ...             }
    ...         ]
    ...     }})
    True
    >>> on_payment_holiday(None, {"loc_record": {
    ...         "modification": [
    ...             {
    ...                 "applied": "2021-07-01T00:00:00.000Z",
    ...                 "suspended": None,
    ...                 "instalment_amount": 0
    ...             }
    ...         ]
    ...     }})
    True
    >>> on_payment_holiday(None, {"loc_record": {
    ...         "modification": [
    ...             {
    ...                 "applied": "2021-07-01T00:00:00.000Z",
    ...                 "suspended": "2021-08-01T00:00:00.000Z",
    ...                 "instalment_amount": 0
    ...             }
    ...         ]
    ...     }})
    False
    """

    # Identify if there are any active payment holiday modifications
    payment_holidays = [
        x
        for x in payload["loc_record"]["modification"]
        if x["instalment_amount"] == 0
        if datetime.strptime(x["applied"], "%Y-%m-%dT%H:%M:%S.%fZ") <= datetime.today()
        if datetime.strptime(
            x["suspended"] or "2070-01-01T00:00:00.000Z", "%Y-%m-%dT%H:%M:%S.%fZ"
        )
        >= datetime.today()
    ]

    # Fail if any active payment holidays are found
    return len(payment_holidays) > 0


@rule(rule_id="PR-004-S1", name="Had Payment Holiday")
def had_payment_holiday(_, payload):
    """
    Supporting rule to indicate whether a an account has previously been on a payment holiday.
    A pass result indicates the account has previously been subject to payment holiday.

    >>> had_payment_holiday(None, {"loc_record": {
    ...         "modification": [
    ...             {
    ...                 "applied": "2021-07-01T00:00:00.000Z",
    ...                 "suspended": "2021-08-01T00:00:00.000Z",
    ...                 "instalment_amount": 0
    ...             }
    ...         ]
    ...     }})
    True
    >>> had_payment_holiday(None, {"loc_record": {
    ...         "modification": [
    ...             {
    ...                 "applied": "2021-07-01T00:00:00.000Z",
    ...                 "suspended": None,
    ...                 "instalment_amount": 0
    ...             }
    ...         ]
    ...     }})
    False
    >>> had_payment_holiday(None, {"loc_record": {
    ...         "modification": [
    ...             {
    ...                 "applied": "2021-07-01T00:00:00.000Z",
    ...                 "suspended": "2021-12-01T00:00:00.000Z",
    ...                 "instalment_amount": 0
    ...             }
    ...         ]
    ...     }})
    False
    """

    # Pass if there are any previous, non-active payment holiday account modifications.
    return (
        len(
            [
                x
                for x in payload["loc_record"]["modification"]
                if x["instalment_amount"] == 0
                if datetime.strptime(x["applied"], "%Y-%m-%dT%H:%M:%S.%fZ")
                < datetime.today()
                if x["suspended"] is not None
                if datetime.strptime(
                    x["suspended"] or "2070-01-01T00:00:00.000Z",
                    "%Y-%m-%dT%H:%M:%S.%fZ",
                )
                < datetime.today()
            ]
        )
        > 0
    )


@rule(
    rule_id="PR-004-U1",
    name="Three Consecutive Payments Since Payment Holiday",
    run_if=had_payment_holiday,
)
def three_consecutive_payments_since_payment_holiday(_, payload):
    """
    Unblock if three consecutive, non-reduced payments have been made since the last payment holiday ended.

    >>> three_consecutive_payments_since_payment_holiday(None, {"loc_record": {"current_cons_full_pmt": 3}})
    True
    >>> three_consecutive_payments_since_payment_holiday(None, {"loc_record": {"current_cons_full_pmt": 2}})
    False
    """

    # Fail if current numebr of consectutive full payments is less then 3.
    return payload["loc_record"]["current_cons_full_pmt"] >= 3


block_on_one_missed_payment = All(
    "MP-001",
    "Block On One Missed Payment",
    days_not_paid,
    two_consecutive_payments_made,
    description="Customer has missed their latest payment.",
)

block_on_two_missed_payments = All(
    "MP-002",
    "Block On Two Missed Payments",
    two_missed_payments,
    three_consecutive_payments_made,
    description="Customer has missed two payments in the last three months.",
)

block_on_three_missed_payments = All(
    "MP-003",
    "Block On Three Missed Payments",
    three_missed_payments,
    four_consecutive_payments_made,
    description="Customer has missed three payments in the last three months.",
)

block_on_more_than_three_missed_payments = All(
    "MP-004",
    "Block On More Than Three Missed Payments",
    more_than_three_missed_payments,
    six_consecutive_payments_made,
    description="Customer has missed more than three payments in the last six months.",
)

missed_payments = All(
    "MP",
    "Missed Payments",
    block_on_one_missed_payment,
    block_on_two_missed_payments,
    block_on_three_missed_payments,
    block_on_more_than_three_missed_payments,
    description="Accounts blocks caused by missing or late payments.",
)


block_on_two_payment_reductions_in_180_days = All(
    "PR-001",
    "Block on Two Payment Reductions in 180 Days",
    two_payment_reductions_in_180_days,
    fifty_percent_principal_paid,
    no_reductions_since_latest_draw,
    description="There have been two payment reductions in the last 180 days.",
)

block_on_two_payment_reductions_in_90_days = All(
    "PR-002",
    "Block on Two Payment Reductions in 90 Days",
    two_payment_reductions_in_90_days,
    seventy_five_percent_principal_paid,
    no_reductions_since_latest_draw,
    description="There have been two payment reductions in the last 90 days.",
)

block_on_two_consecutive_payment_reductions = All(
    "PR-003",
    "Block on Two Consecutive Payment Reductions",
    two_consecutive_payment_reductions,
    seventy_five_percent_principal_paid,
    no_reductions_since_latest_draw,
    description="There have been two payment reductions in the last 90 days.",
)

block_on_payment_holiday = All(
    "PR-004",
    "Block on Payment Holiday",
    on_payment_holiday,
    three_consecutive_payments_since_payment_holiday,
    description="The account is subject to an agreed payment holiday.",
)

payment_reductions = All(
    "PR",
    "Payment Reductions",
    block_on_two_payment_reductions_in_180_days,
    block_on_two_payment_reductions_in_90_days,
    block_on_two_consecutive_payment_reductions,
    block_on_payment_holiday,
    description="Accounts blocks caused by multiple payment reductions.",
)


account_status = All(
    "account_status",
    "Account Status",
    missed_payments,
    payment_reductions,
    description="Overall status for the account, where failure indicates that the account should be blocked.",
)

nest_10 = All(
    "nest_10",
    "Nest 10 this name is super super super super super super super super super super long",
    days_not_paid,
    description="Nest stress test",
)

nest_9 = All(
    "nest_9",
    "Nest 9",
    nest_10,
    description="Nest stress test",
)

nest_8 = All(
    "nest_8",
    "Nest 8",
    nest_9,
    description="Nest stress test",
)

nest_7 = All(
    "nest_7",
    "Nest 7 this name is very descriptive and probably too long to be useful",
    nest_8,
    description="Nest stress test",
)

nest_6 = All(
    "nest_6",
    "Nest 6 short name again",
    nest_7,
    description="Nest stress test",
)

nest_5 = All(
    "nest_5",
    "Nest 5 This name is very long, even longer than the last",
    nest_6,
    description="Nest stress test",
)

nest_4 = All(
    "nest_4",
    "Nest 4 an even longer name here to fill space",
    nest_5,
    description="Nest stress test",
)

nest_3 = All(
    "nest_3",
    "Nest 3 middle length name",
    nest_4,
    description="Nest stress test",
)

nest_2 = All(
    "nest_2",
    "Nest 2 short name",
    nest_3,
    description="Nest stress test",
)

nest_1 = All(
    "nest_1",
    "Nest 1",
    nest_2,
    description="Nest stress test",
)
